/*
 * nsVanadielAlarm.js
 *
 * Copyright (c) 2005, 2006 by Daniel Potter.  All rights reserved.
 *
 * This file is part of the Vana'diel Clock.
 *
 * The Vana'diel Clock is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The Vana'diel Clock is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Vana'diel Clock; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Manages alarms.
 */

// From nsIPrefBranch2.idl
const NS_PREFBRANCH_PREFCHANGE_TOPIC_ID = "nsPref:changed";

// Constants
const VANADIEL_CLOCK_NAMESPACE = "http://ffxi.xenoveritas.org/tools/vanaclock/rdf#";

const VERSION_PROPERTY  = VANADIEL_CLOCK_NAMESPACE + "version";
const ALARM_PROPERTY    = VANADIEL_CLOCK_NAMESPACE + "alarm";
const USER_ALARM_PREFIX = VANADIEL_CLOCK_NAMESPACE + "user-alarm-";

const ALARM_PREFIX = "http://ffxi.xenoveritas.org/tools/vanaclock/rdf#alarm-";

const ALARM_NAME_PROPERTY        = ALARM_PREFIX + "name";
const ALARM_DESCRIPTION_PROPERTY = ALARM_PREFIX + "description";
const ALARM_PATTERN_PROPERTY     = ALARM_PREFIX + "pattern";
const ALARM_TIME_PROPERTY        = ALARM_PREFIX + "time";
const ALARM_TIME_BEFORE_PROPERTY = ALARM_PREFIX + "time-before";
const ALARM_ICON_PROPERTY        = ALARM_PREFIX + "icon";
const ALARM_SMALL_ICON_PROPERTY  = ALARM_PREFIX + "icon-small";
const ALARM_SOUND_PROPERTY       = ALARM_PREFIX + "sound";
const ALARM_TYPE_PROPERTY        = ALARM_PREFIX + "type";

const DAY_ICON_PATH = "chrome://vanadielclock/skin/elements/";
const MOON_ICON_PATH = "chrome://vanadielclock/skin/moon/";

const ICON_EXTENSION = ".png";

const DAY_ICONS = [ "fire", "earth", "water", "wind", "ice", "thunder", "light", "dark" ];
const MOON_ICONS = [ "new", "waxcres1", "waxcres2", "firstquarter", "waxgibb1", "waxgibb2", "full", "wanegibb1", "wanegibb2", "lastquarter", "wanecres1", "wanecres2" ];

var lastAlarmGUID = 0;

function VanadielAlarm() {
    var stamp = Math.floor(new Date().getTime() / 1000);
    if (stamp == lastAlarmGUID) {
	stamp++;
    }
    lastAlarmGUID = stamp;
    this.mId = "A" + stamp;
}

VanadielAlarm.prototype = {
    RDF: Components.classes["@mozilla.org/rdf/rdf-service;1"]
	       .getService(Components.interfaces.nsIRDFService),
    mId: null,
    mNextAlarm: 0,
    mTime: 0,
    mPreAlertTime: 0,
    mPattern: 0,
    mName: "",
    mDescription: "",
    mIcon: "",
    mSmallIcon: "",
    mSound: "",
    mType: "",
    mDebugMsg: "",
    mRDFSource: null,
    mRDFNode: null,
    
    get id() { return this.mId; },
    
    get nextAlarm() { return this.mNextAlarm > 0 ? this.mNextAlarm : 0; },
    
    get active() {
	return this.mNextAlarm != -1;
    },

    get pattern() { return this.mPattern; },
    set pattern(value) {
	if (value == null)
	    value = "";
	this.mPattern = value;
	this.setRDF(ALARM_PATTERN_PROPERTY, value);
    },

    get name() { return this.mName; },
    set name(value) {
	if (value != null) {
	    this.mName = value;
	    this.setRDF(ALARM_NAME_PROPERTY, value);
	}
    },

    get description() { return this.mDescription; },
    set description(value) {
	if (value != null) {
	    this.mDescription = value;
	    this.setRDF(ALARM_DESCRIPTION_PROPERTY, value);
	}
    },

    get time() { return this.mTime; },
    set time(value) {
	if (value != null) {
	    this.mTime = value;
	    this.mNextAlarm = this.mTime - this.mPreAlertTime;
	    this.setRDF(ALARM_TIME_PROPERTY, value);
	}
    },

    get preAlertTime() { return this.mPreAlertTime; },
    set preAlertTime(value) {
	if (value != null) {
	    this.mPreAlertTime = value;
	    this.mNextAlarm = this.mTime - this.mPreAlertTime;
	    this.setRDF(ALARM_TIME_BEFORE_PROPERTY, value);
	}
    },

    get icon() { return this.mIcon; },
    set icon(value) {
	if (value != null) {
	    this.mIcon = value;
	    this.setRDF(ALARM_ICON_PROPERTY, value);
	}
    },

    get nextIcon() {
	if (this.mIcon == "time:moon") {
	    return MOON_ICON_PATH + MOON_ICONS[this.getNextPhase()] + ICON_EXTENSION;
	} else if (this.mIcon == "time:day") {
	    return DAY_ICON_PATH + DAY_ICONS[this.getNextDay()] + "32" + ICON_EXTENSION;
	}
	return this.mIcon;
    },

    get smallIcon() { return this.mSmallIcon; },
    set smallIcon(value) {
	if (value != null) {
	    this.mSmallIcon = value;
	    this.setRDF(ALARM_SMALL_ICON_PROPERTY, value);
	}
    },

    get nextSmallIcon() {
	if (this.mSmallIcon == "time:moon") {
	    return MOON_ICON_PATH + MOON_ICONS[this.getNextPhase()] + ICON_EXTENSION;
	} else if (this.mSmallIcon == "time:day") {
	    return DAY_ICON_PATH + DAY_ICONS[this.getNextDay()] + ICON_EXTENSION;
	}
	return this.mSmallIcon;
    },

    get sound() { return this.mSound; },
    set sound(value) {
	if (value != null) {
	    this.mSound = value;
	    this.setRDF(ALARM_SOUND_PROPERTY, value);
	}
    },
    
    get type() { return this.mType; },
    set type(value) {
	if (value != null) {
	    this.mType = value;
	    this.setRDF(ALARM_TYPE_PROPERTY, value);
	}
    },
    
    get debugMessage() { return this.mDebugMsg; },

    setAlarm: function(time, pattern) {
	this.mTime = this.calculateTime(time, pattern);
	this.mNextAlarm = this.mTime - this.mPreAlertTime;
    },
    
    updateAlarm: function(time) {
	dump("update alarm\n");
	this.setAlarm(time, this.mPattern);
    },
    
    /**
     * Returns the day on which this sounds (as a number).
     */
    getNextDay: function() {
	return Math.floor(this.mTime / 86400) % 8;
    },
    
    /**
     * Returns the moon phase on which this sounds (as a number).
     */
    getNextPhase: function() {
	return Math.floor((this.mTime + 6220800) / 604800) % 12;
    },

    /**
     * Calculates the next time the pattern will be hit based on
     * the current time.
     */
    calculateTime: function(time, pattern) {
	// Parse the pattern
	var tokens = pattern.split(" ");
	var jsPat = new Array(5);
	// this is probably unnecessary, but:
	for (var i = 0; i < 5; i++) {
	    jsPat[i] = null;
	}
	for (var t = 0; t < tokens.length; t++) {
	    var tok = new Object();
	    // If this is set to * ignore it
	    if (tokens[t] == "*")
		continue;
	    if (t > 4) {
		// at one point there was going to be extended syntax
		// right now there isn't
		break;
	    }
	    tok.offset = 0;
	    switch (t) {
		case 0:
		    // Weekdays
		    tok.period = 86400;
		    tok.cycles = 8;
		    jsPat[2] = tok;
		    break;
		case 1:
		    // hours
		    tok.period = 3600;
		    tok.cycles = 24;
		    jsPat[3] = tok;
		    break;
		case 2:
		    // minutes
		    tok.period = 60;
		    tok.cycles = 60;
		    jsPat[4] = tok;
		    break;
		case 3:
		    // moonphase
		    tok.period = 604800;
		    tok.cycles = 12;
		    tok.offset = 6220800;
		    jsPat[0] = tok;
		    break;
		case 4:
		    // rse
		    tok.period = 691200;
		    tok.cycles = 8;
		    jsPat[1] = tok;
		    break;
	    }
	    // now calculate the token matches
	    if (tokens[t].indexOf(",") >= 0) {
		var matches = tokens[t].split(",");
		var i;
		tok.matches = new Array();
		for (var m = 0; m < matches.length; m++) {
		    i = parseInt(matches[m]);
		    if (!(i >= 0))
			i = 0;
		    if (i >= tok.cycles)
			i = tok.cycles - 1;
		    tok.matches.push(i);
		}
	    } else if (tokens[t].indexOf("-") >= 0) {
		var i = tokens[t].indexOf("-");
		var start = parseInt(tokens[t].substring(0, i));
		var end = parseInt(tokens[t].substring(i+1));
		// The following check for NaN, as comparisons against NaN
		// are always false
		if (!(start >= 0))
		    start = 0;
		if (!(end > start))
		    end = start;
		if (end >= tok.cycles)
		    end = tok.cycles - 1;
		// Catch totally invalid values
		if (start > end)
		    start = end;
		// sweep through and dump them in
		tok.matches = new Array(end - start + 1);
		for (i = 0; start <= end; i++,start++) {
		    tok.matches[i] = start;
		}
	    } else {
		tok.matches = new Array(1);
		tok.matches[0] = parseInt(tokens[t]);
		if (!(tok.matches[0] >= 0))
		    tok.matches[0] = 0;
		if (tok.matches[0] >= tok.cycles)
		    tok.matches[0] = tok.cycles - 1;
	    }
	}
	/*
	this.mDebugMsg = "";
	this.debug("Pattern parsed: ");
	for (var i in jsPat) {
	    if (jsPat[i] == null) {
		this.debug("*");
	    } else {
		for (var j = 0; j < jsPat[i].matches.length; j++) {
		    if (j > 0)
			this.debug(",");
		    this.debug(jsPat[i].matches[j]);
		}
	    }
	    this.debug(" ");
	}
	this.debug("\n");
	*/
	// now do an actual calculate sweep:
	var nTime = time;
	for (var i = 0; i < 5; i++) {
	    var sweep = this.calculateTimeSweep(time, nTime, Infinity, jsPat, 0);
	    if (sweep == null) {
		dump("Pattern " + pattern + " never matched anything!  Setting alarm to go off a year later!\n");
		return nTime + 360 * 24 * 60 * 60;
	    }
	    if (sweep.earliest > time) {
		//dump("Calculated pattern " + pattern + " to go off at " + sweep.earliest + " from " + time + "\n");
		return sweep.earliest;
	    }
	    // move to (just) after the sweep we got.
	    nTime = sweep.latest + 1;
	}
	dump("Gave up on " + pattern + " after 5 tries!  Setting alarm to go off a year later!\n");
	return time + 360 * 24 * 60 * 60;
    },
    
    /**
     * calculates a time sweep, returning an Object that contains
     * two entries, "earliest" and "latest" that indicate the span
     * the current pattern and all child entries after it can accept
     *
     * if there's no time it can accept, it returns null
     *
     * after    - any generated time must be after this value
     * earliest - earliest time acceptable
     * latest   - latest time acceptable
     * patterns - array of pattern objects
     * index    - current pattern being checked
     */
    calculateTimeSweep: function(after, earliest, latest, patterns, index) {
	while (patterns[index] == null) {
	    index++;
	    if (index >= patterns.length) {
		// No more patterns, gen this range and let it trickle up
		var res = new Object();
		res.earliest = earliest;
		res.latest = latest;
		return res;
	    }
	}
	var pattern = patterns[index];
	// this.debug("Constraining span [" + earliest + "-" + latest + "], period=" + pattern.period + "; cycles=" + pattern.cycles + "\n");
	var myEarliest, myLatest, starter, span, res;
	// calculate the next pattern we match
	for (var p = 0; p < 10; p++) {
	    for (var i = 0; i < pattern.matches.length; i++) {
		// calculate the start of this current period
		span = pattern.period * pattern.cycles;
		starter = earliest - ((earliest + pattern.offset) % span) + pattern.matches[i] * pattern.period;
		// Is this span within the current span?
		myLatest = starter + pattern.period;
		// this.debug(" - Calculated my span as [" + starter + "," +myLatest + "], ");
		if (myLatest > earliest && starter < latest) {
		    // See if this span is acceptable to our children
		    myEarliest = starter > earliest ? starter : earliest;
		    if (myLatest > latest)
			myLatest = latest;
		    // this.debug("constrainted to [" + myEarliest + "," +myLatest + "].\n");
		    res = this.calculateTimeSweep(after, myEarliest, myLatest, patterns, index + 1);
		    if (res != null && res.earliest > after) {
			// Match was OK, go with that
			return res;
		    }
		} else {
		    // this.debug("which is out of range.\n");
		}
		// Otherwise, continue on to the next match we can accept
	    }
	    // not acceptable, jump ahead to the next span
	    /*
	    this.debug("Failed all matches, moving from ");
	    this.debug(earliest);
	    this.debug(" to ");
	    */
	    earliest = earliest - ((earliest + pattern.offset) % span) + span;
	    // this.debug(earliest);
	    if (earliest > latest) {
		// this.debug(", which bumped us out of the current range.\n");
		// not OK, fail this
		return null;
	    }
	    // this.debug(".\n");
	}
	dump(" ************* VANA'DIEL CLOCK PATTERN MATCH OVER TEN PASSES!\n");
	dump(" ************* THIS SHOULD NEVER HAPPEN.\n");
	// if we drop out, none of our times were acceptable, so we return null
	return null;
    },
    
    /*
    debug: function(msg) {
	this.mDebugMsg = this.mDebugMsg + msg;
    },
    */

    activate: function(time) {
	// move us to the next alarm occurance
	if (time < this.mTime) {
	    this.mNextAlarm = this.mTime - this.mPreAlertTime;
	} else {
	    // ok, next alarm then
	    if (this.mPattern == "") {
		this.mNextAlarm = -1;
		return false;
	    }
	    this.mTime = this.calculateTime(time, this.mPattern);
	    this.mNextAlarm = this.mTime - this.mPreAlertTime;
	    // save the alarm time
	    this.setRDF(ALARM_TIME_PROPERTY, this.mNextAlarm);
	}
	if (this.mNextAlarm <= time)
	    this.mNextAlarm = this.mTime;
	return true;
    },
    
    /**
     * Loads the information from the given rdfResource.
     * This may reset the autogenerated ID.
     */
    bindToRDF: function(rdfSource, rdfResource) {
	this.mRDFSource = rdfSource;
	this.mRDFNode = rdfResource;
	if (rdfResource.Value.substring(0, USER_ALARM_PREFIX.length) == USER_ALARM_PREFIX) {
	    dump("ID changed from ");
	    dump(this.mId);
	    this.mId = rdfResource.Value.substring(USER_ALARM_PREFIX.length);
	    dump(" to ");
	    dump(this.mId);
	    dump(".\n");
	}
	this.mName         =          this.getOrSetRDF(ALARM_NAME_PROPERTY,        this.mName);
	this.mDescription  =          this.getOrSetRDF(ALARM_DESCRIPTION_PROPERTY, this.mDescription);
	this.mPattern      =          this.getOrSetRDF(ALARM_PATTERN_PROPERTY,     this.mPattern);
	this.mTime         = parseInt(this.getOrSetRDF(ALARM_TIME_PROPERTY,        this.mTime));
	this.mPreAlertTime = parseInt(this.getOrSetRDF(ALARM_TIME_BEFORE_PROPERTY, this.mPreAlertTime));
	this.mIcon         =          this.getOrSetRDF(ALARM_ICON_PROPERTY,        this.mIcon);
	this.mSmallIcon    =          this.getOrSetRDF(ALARM_SMALL_ICON_PROPERTY,  this.mSmallIcon);
	this.mSound        =          this.getOrSetRDF(ALARM_SOUND_PROPERTY,       this.mSound);
	this.mType         =          this.getOrSetRDF(ALARM_TYPE_PROPERTY,        this.mType);
	// correct NaN errors
	if (!(this.mTime >= 0)) {
	    this.mTime = 0;
	}
	if (!(this.mPreAlertTime >= 0)) {
	    this.mPreAlertTime = 0;
	}
	this.mNextAlarm = this.mTime - this.mPreAlertTime;
	this.dumpAlarm();
    },
    
    unbindFromRDF: function() {
	if (this.mRDFSource == null)
	    return;
	// unassert everything
	this.unsetRDF(ALARM_NAME_PROPERTY);
	this.unsetRDF(ALARM_DESCRIPTION_PROPERTY);
	this.unsetRDF(ALARM_PATTERN_PROPERTY);
	this.unsetRDF(ALARM_TIME_PROPERTY);
	this.unsetRDF(ALARM_TIME_BEFORE_PROPERTY);
	this.unsetRDF(ALARM_ICON_PROPERTY);
	this.unsetRDF(ALARM_SMALL_ICON_PROPERTY);
	this.unsetRDF(ALARM_SOUND_PROPERTY);
	this.unsetRDF(ALARM_TYPE_PROPERTY);
	this.mRDFSource = null;
	this.mRDFNode = null;
    },
    
    getOrSetRDF: function(name, value) {
	var res = this.RDF.GetResource(name);
	var v = this.mRDFSource.GetTarget(this.mRDFNode, res, true);
	if (v == null) {
	    // set it
	    this.mRDFSource.Assert(this.mRDFNode, res, this.RDF.GetLiteral(value), true);
	    return value;
	} else {
	    return v.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
	}
    },

    setRDF: function(name, value) {
	// allow sets if we aren't bound...
	if (this.mRDFSource == null)
	    return;
	var res = this.RDF.GetResource(name);
	var v = this.mRDFSource.GetTarget(this.mRDFNode, res, true);
	if (v == null) {
	    // set it
	    this.mRDFSource.Assert(this.mRDFNode, res, this.RDF.GetLiteral(value), true);
	} else {
	    // change it
	    this.mRDFSource.Change(this.mRDFNode, res, v, this.RDF.GetLiteral(value));
	}
    },
    
    unsetRDF: function(name) {
	var res = this.RDF.GetResource(name);
	var v = this.mRDFSource.GetTarget(this.mRDFNode, res, true);
	if (v != null) {
	    // unassert it
	    this.mRDFSource.Unassert(this.mRDFNode, res, v);
	}
    },

    QueryInterface: function (iid) {
        if (!iid.equals(Components.interfaces.nsIVanadielAlarm)
            && !iid.equals(Components.interfaces.nsISupports))
        {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    },
    
    dumpAlarm: function () {
	dump("Alarm:\n");
	dump("            id: " + this.mId + "\n");
	dump("     nextAlarm: " + this.mNextAlarm + "\n");
	dump("          name: " + this.mName + "\n");
	dump("   description: " + this.mDescription + "\n");
	dump("       pattern: " + this.mPattern + "\n");
	dump("          time: " + this.mTime + "\n");
	dump("  preAlertTime: " + this.mPreAlertTime + "\n");
	dump("          icon: " + this.mIcon + "\n");
	dump("     smallIcon: " + this.mSmallIcon + "\n");
	dump("         sound: " + this.mSound + "\n");
	dump("          type: " + this.mType + "\n");
    }
}

function VanadielAlarmManager() {
    this.alarms = new Array();
    var profileInternal = Components.classes['@mozilla.org/file/directory_service;1']
                                    .getService(Components.interfaces.nsIProperties);
    var file = profileInternal.get('ProfD', Components.interfaces.nsILocalFile);
    file.append("vanadiel-alarms.rdf");
    var fileHandler = Components.classes["@mozilla.org/network/io-service;1"]
                        .getService(Components.interfaces.nsIIOService);
    var url = fileHandler.newFileURI(file);
    this.mRdfUri = url.spec;
    dump("loading alarms from " + this.mRdfUri + "\n");
    try {
	this.alarmDS = this.RDF.GetDataSourceBlocking(this.mRdfUri);
	// see if there's a version already set
	this.clockNS = this.RDF.GetResource(VANADIEL_CLOCK_NAMESPACE);
	this.alarmNS = this.RDF.GetResource(ALARM_PROPERTY);
	var versionNS = this.RDF.GetResource(VERSION_PROPERTY);
	var version = this.alarmDS.GetTarget(this.clockNS, versionNS, true);
	if (version == null) {
	    // No - set the version
	    this.alarmDS.Assert(this.clockNS, versionNS, this.RDF.GetLiteral("1.1"), true);
	} else if (version.QueryInterface(Components.interfaces.nsIRDFLiteral).Value == "1.0") {
	    // Mark for "upgrade" to 1.1 (1.1 alters moonphase requirements)
	    // Not exactly clear how the upgrade is going to work, but...
	}
	// Try and get the alarm RDF sequence
	if (this.RDFCUtils.IsSeq(this.alarmDS, this.alarmNS)) {
	    this.alarmRDFSeq = Components.classes['@mozilla.org/rdf/container;1']
				    .createInstance(Components.interfaces.nsIRDFContainer);
	    this.alarmRDFSeq.Init(this.alarmDS, this.alarmNS);
	} else {
	    // create it
	    this.alarmRDFSeq = this.RDFCUtils.MakeSeq(this.alarmDS, this.alarmNS);
	}
	this.alarmDS.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
    } catch (ex) {
	// TODO: send JS errors to the JavaScript console?
	dump("Failed to load alarm from RDF:\n");
	dump(ex);
	dump("\n");
	for (var key in ex) {
	    dump("exception.");
	    dump(key);
	    dump(": ");
	    dump(ex[key]);
	    dump("\n");
	}
    }
    this.mListeners = new Array();
    var prefserv = Components.classes["@mozilla.org/preferences;1"].
	    getService(Components.interfaces.nsIPrefBranch2);
    prefserv.addObserver("vanadiel.clock.", this, false);
    prefserv = prefserv.QueryInterface(Components.interfaces.nsIPrefBranch);
    this.mEnabled = prefserv.getBoolPref("vanadiel.clock.alarm.enable");
    this.mType = prefserv.getCharPref("vanadiel.clock.alarm.announce");
}

VanadielAlarmManager.prototype = {
    alarms: null,
    RDF: Components.classes["@mozilla.org/rdf/rdf-service;1"]
	    .getService(Components.interfaces.nsIRDFService),
    RDFCUtils: Components.classes["@mozilla.org/rdf/container-utils;1"]
		.getService(Components.interfaces.nsIRDFContainerUtils),
    alarmDS: null,
    clockNS: null,
    alarmNS: null,
    alarmRDFSeq: null,
    loaded: false,
    mRdfUri: "",
    mListeners: null,
    mEnabled: true,
    mType: "popup",
    
    get rdfURI() { return this.mRdfUri; },
    
    get count() { return this.alarms.length; },
    
    getAlarmAt: function(index) {
	if (index < 0 || index >= this.alarms.length)
	    return null;
	return this.alarms[index];
    },
    
    getAlarmById: function(id) {
	for (var a in this.alarms) {
	    if (this.alarms[a].id == id)
		return this.alarms[a];
	}
	return null;
    },
    
    addAlarm: function(alarm) {
	// See if this alarm is in the array
	var index;
	for (index = 0; index < this.alarms.length; index++) {
	    if (this.alarms[index] == alarm) {
		// do nothing
		return;
	    }
	}
	this.alarms.push(alarm);
	// add the alarm to our datasource (creating a GUID for this alarm)
	var alarmNode = this.RDF.GetResource(USER_ALARM_PREFIX + alarm.id);
	this.alarmRDFSeq.AppendElement(alarmNode);
	alarm.bindToRDF(this.alarmDS, alarmNode);
	this.flushAlarms();
	// alert listeners that an alarm has come into existance
	for (var l in this.mListeners) {
	    try {
		this.mListeners[l].alarmAdded(alarm);
	    } catch (ex) {
		// alert on error
		// TODO: send JS errors to the JavaScript console?
		dump("Listener caused exception:\n");
		dump(ex);
		dump("\n");
		for (var key in ex) {
		    dump("exception.");
		    dump(key);
		    dump(": ");
		    dump(ex[key]);
		    dump("\n");
		}
	    }
	}
    },
    
    removeAlarm: function(alarm) {
	// Find this alarm in the array
	var index;
	for (index = 0; index < this.alarms.length; index++) {
	    if (this.alarms[index] == alarm) {
		// remove the alarm
		var node = this.RDF.GetResource(USER_ALARM_PREFIX + alarm.id);
		this.alarmRDFSeq.RemoveElement(node, true);
		alarm.unbindFromRDF();
		// remove the alarm from the array
		this.alarms.splice(index, 1);
		this.flushAlarms();
		// alert listeners that alarms have vanished
		for (var l in this.mListeners) {
		    try {
			this.mListeners[l].alarmRemoved(alarm);
		    } catch (ex) {
			// alert on error
			// TODO: send JS errors to the JavaScript console?
			dump("Listener caused exception:\n");
			dump(ex);
			dump("\n");
			for (var key in ex) {
			    dump("exception.");
			    dump(key);
			    dump(": ");
			    dump(ex[key]);
			    dump("\n");
			}
		    }
		}
		break;
	    }
	}
    },
    
    loadAlarms: function() {
	if (this.loaded)
	    return;
	this.loaded = true;
	var elements = this.alarmRDFSeq.GetElements();
	this.alarms = new Array();
	// go through the alarms and create them
	while (elements.hasMoreElements()){
	    dump("Collecting alarm\n");
	    var alarmRes = elements.getNext();
	    if (alarmRes instanceof Components.interfaces.nsIRDFResource) {
		// create this alarm
		var alarm = Components.classes["@xenoveritas.org/vanadiel-alarm;1"].
			   createInstance(Components.interfaces.nsIVanadielAlarm);
		try {
		    alarm.bindToRDF(this.alarmDS, alarmRes);
		    this.alarms.push(alarm);
		} catch (ex) {
		    // TODO: send JS errors to the JavaScript console?
		    dump("Error loading alarm \"");
		    dump(alarmRes.Value);
		    dump("\":\n");
		    dump(ex);
		    dump("\n");
		    for (var key in ex) {
			dump("exception.");
			dump(key);
			dump(": ");
			dump(ex[key]);
			dump("\n");
		    }
		}
	    }
	}
    },
    
    flushAlarms: function() {
	this.alarmDS.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
    },
    
    activateAlarms: function(time, late) {
	var alarmsSounded = null;
	var alarmSound = "none";
	for (var a in this.alarms) {
	    if (this.alarms[a].nextAlarm <= time) {
		if (alarmsSounded == null) {
		    alarmsSounded = new Array();
		}
		alarmsSounded.push(this.alarms[a]);
		// "upgrade" the alarm sound, if possible
		if (alarmSound == "none") {
		    alarmSound = this.alarms[a].sound;
		} else if (alarmSound == "beep") {
		    if (this.alarms[a].sound != "none") {
			alarmSound = this.alarms[a].sound;
		    }
		}
	    }
	}
	if (alarmsSounded == null) {
	    // nothing has changed
	    return;
	} else {
	    // Alarms still fire when we're disabled, we just don't display a
	    // dialog box when disabled.
	    if (this.mEnabled) {
		this.showAlarmDialog(alarmsSounded);
	    }
	    // activate all the alarms after the dialog has been spawned
	    for (var a in alarmsSounded) {
		alarmsSounded[a].activate(time);
		// and alert listeners
		for (var l in this.mListeners) {
		    try {
			this.mListeners[l].alarmFired(alarm);
		    } catch (ex) {
			// chomp-chomp
		    }
		}
	    }
	    if (this.mEnabled) {
		if (alarmSound != "none") {
		    var soundService = Components.classes["@mozilla.org/sound;1"].createInstance(Components.interfaces.nsISound);
		    soundService.init();
		    if (alarmSound == "beep")
			soundService.beep();
		    else if (alarmSound.substring(0, 6) == "chrome")
			soundService.play(alarmSound);
		    else
			soundService.playSystemSound(alarmSound);
		}
	    }
	}
	//var alertService = Components.classes["@mozilla.org/alerts-service;1"].getService(Components.interfaces.nsIAlertsService);
	//alertService.showAlertNotification(alarmIcon, alarmTitle, alarmDesc, false, "don't care", null);

	// see if we need to remove any alarms and kill them
	var a = 0;
	while (a < this.alarms.length) {
	    if (!this.alarms[a].active) {
		this.removeAlarm(this.alarms[a]);
		continue;
	    }
	    a++;
	}
	this.flushAlarms();
	this.fireAlarmAltered();
    },

    fireAlarmAltered: function() {
	for (var l in this.mListeners) {
	    try {
		this.mListeners[l].alarmsAltered();
	    } catch (ex) {
		dump("Listener caused exception:\n");
		dump(ex);
		dump("\n");
		for (var key in ex) {
		    dump("exception.");
		    dump(key);
		    dump(": ");
		    dump(ex[key]);
		    dump("\n");
		}
	    }
	}
    },

    showAlarmDialog: function(alarm) {
	// this can either be an array of alarms or a single alarm
	var alarms;
	if ("name" in alarm) {
	    // single alarm
	    alarms = new Array(1);
	    alarms[0] = alarm;
	} else {
	    alarms = alarm;
	}
	var args = new Array(alarms.length);
	for (var i = 0; i < alarms.length; i++) {
	    args[i] = new Object();
	    // copy over the key fields, since the alarm object is about to
	    // be activated, which may change them
	    args[i].name = alarms[i].name;
	    args[i].description = alarms[i].description;
	    args[i].icon = alarms[i].nextIcon;
	}
	var appShellServ = Components.classes["@mozilla.org/appshell/appShellService;1"]
		.getService(Components.interfaces.nsIAppShellService);
	var hiddenWin = appShellServ.hiddenDOMWindow;
	var popup = this.mType == "popup";
	var style = popup ?
		'chrome,dialog=yes,titlebar=no,popup=yes' :
		'chrome,all,dialog=no';
	hiddenWin
		.openDialog('chrome://vanadielclock/content/alarm-announce.xul',
		'_blank', style, args, !popup, popup);
    },
    
    // Check for preference changes
    observe: function(subject, topic, data) {
	if (topic == NS_PREFBRANCH_PREFCHANGE_TOPIC_ID) {
	    var branch = subject.QueryInterface(Components.interfaces.nsIPrefBranch);
	    if (data == "vanadiel.clock.alarm.enable") {
		this.mEnabled = branch.getBoolPref("vanadiel.clock.alarm.enable");
	    } else if (data == "vanadiel.clock.alarm.announce") {
		this.mType = branch.getCharPref("vanadiel.clock.alarm.announce");
	    }
	    // otherwise, we don't care!
	}
    },

    addVanadielAlarmManagerListener: function(listener) {
	for (var l in this.mListeners) {
	    if (this.mListeners[l] == listener)
		return;
	}
	this.mListeners.push(listener);
	//dump("Vana'diel Alarm Manager: Added listener, up to " + this.mListeners.length + " listeners\n");
    },

    removeVanadielAlarmManagerListener: function(listener) {
	for (var l = 0; l < this.mListeners.length; l++) {
	    if (this.mListeners[l] == listener) {
		this.mListeners.splice(l, 1);
		//dump("Vana'diel Alarm Manager: Removed listener, down to " + this.mListeners.length + " listeners\n");
		return;
	    }
	}
	dump("Vana'diel Alarm Manager: Couldn't removed listener, listener does not exist in list\n");
    },

    QueryInterface: function (iid) {
        if (!iid.equals(Components.interfaces.nsIVanadielAlarmManager)
	    && !iid.equals(Components.interfaces.nsIObserver)
            && !iid.equals(Components.interfaces.nsISupports))
        {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    }
}

/**********************************************************
   XPCOM registration stuff, from
   http://books.mozdev.org/html/mozilla-chp-8-sect-2.html
   Modified to use an array of the several components we offer
 **********************************************************/

var Module = {
    firstTime: true,

    registerSelf: function (compMgr, fileSpec, location, type) {
        if (this.firstTime) {
            this.firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        compMgr =
	    compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
	for (var m in this.modules) {
	    compMgr.registerFactoryLocation(this.modules[m].mCID,
                                        this.modules[m].mName,
                                        this.modules[m].mProgID,
                                        fileSpec,
                                        location,
                                        type);
	}
    },

    getClassObject : function (compMgr, cid, iid) {
        if (!iid.equals(Components.interfaces.nsIFactory))
	    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	for (var m in this.modules) {
	    if (cid.equals(this.modules[m].mCID))
		return this.modules[m].mFactory;
	}
	throw Components.results.NS_ERROR_NO_INTERFACE;
    },

    modules: {
	vanadielAlarm: {
	    mCID: Components.ID("{e8513bde-bad9-4172-a67f-8de617a55a57}"),
	    mProgID: "@xenoveritas.org/vanadiel-alarm;1",
	    mName: "Vana'diel Alarm",
	
	    mFactory: {
		createInstance: function (outer, iid) {
		    if (outer != null)
			throw Components.results.NS_ERROR_NO_AGGREGATION;
		    return (new VanadielAlarm()).QueryInterface(iid);
		}
	    },
	},
	vanadielAlarmManager: {
	    mCID: Components.ID("{e8513bde-bad9-4172-a67f-8de617a55a56}"),
	    mProgID: "@xenoveritas.org/vanadiel-alarm-manager;1",
	    mName: "Vana'diel Alarm Manager",
	
	    mFactory: {
		createInstance: function (outer, iid) {
		    if (outer != null)
			throw Components.results.NS_ERROR_NO_AGGREGATION;
		    return (new VanadielAlarmManager()).QueryInterface(iid);
		}
	    },
	}
    },

    canUnload: function(compMgr) {
        return true;
    }
}; // END Module

function NSGetModule(compMgr, fileSpec) { return Module; }
